// ==========================================================================
// $Source: /users/huron/gautier/projets/CVS/Givaro/src/kernel/integer/givintgcd.C,v $
// Copyright(c)'94-97 by Givaro Team
// see the copyright file.
// Authors: M. Samama, T. Gautier
// $Id: givintgcd.C,v 1.0.0.1 1999/04/15 10:13:49 gautier Exp $
// ==========================================================================
// Description: 

#include "givinteger.h"
#include "givcompgmp.h"


// ==========================================================================
// Computes and returns the gcd of the two integers a and b.
// The algorithm used is this of Gmp.
Integer  Integer::GMPGcd(const Integer& a, const Integer& b)
                                  //Integer& u, Integer& v)
{
  Integer Res(Integer::one);
  mpz_gcd( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&(a.gmp_rep), (mpz_ptr)&(b.gmp_rep) ) ;
  if (Res.priv_sign() <0) return -Res;
  return Res ;
}

Integer&  Integer::GMPGcd(Integer& g, const Integer& a, const Integer& b)
                                  //Integer& u, Integer& v)
{
  mpz_gcd( (mpz_ptr)&(g.gmp_rep), (mpz_ptr)&(a.gmp_rep), (mpz_ptr)&(b.gmp_rep) ) ;
  if (g.priv_sign() <0) return negin(g);
  return g ;
}


// ==========================================================================
// Computes and returns the gcd of the two integers a and b.
// Switch to different methods along the size of the input arguments
Integer  gcd(const Integer& a, const Integer& b)
{
  if (&a == &b) return a ;
  return Integer::GMPGcd(a,b) ;
}

Integer&  gcd(Integer& g, const Integer& a, const Integer& b)
{
  if (&a == &b) return g=a ;
  return Integer::GMPGcd(g,a,b) ;
}


// ==========================================================================
// Computes and returns the gcd g of the two integers a and b such that
// g = a*u + b*v .
// The algorithm used is this of Gmp.
Integer  gcd (const Integer& a, const Integer& b, Integer& u, Integer& v)
{
  v = 1; // v must not be 0 to be computed.
  Integer Res(Integer::one);
  mpz_gcdext( (mpz_ptr)&(Res.gmp_rep), (mpz_ptr)&(u.gmp_rep), (mpz_ptr)&(v.gmp_rep), 
              (mpz_ptr)&(a.gmp_rep), (mpz_ptr)&(b.gmp_rep) ) ;
  if (Res.priv_sign() < 0) { Integer::negin(u); Integer::negin(v); return Integer::negin(Res);}
//   { u = -u ; v = -v ; return -Res;}
  return Res;
}

Integer&  gcd (Integer& g, const Integer& a, const Integer& b, Integer& u, Integer& v)
{
  v = 1; // v must not be 0 to be computed.
  mpz_gcdext( (mpz_ptr)&(g.gmp_rep), (mpz_ptr)&(u.gmp_rep), (mpz_ptr)&(v.gmp_rep), 
              (mpz_ptr)&(a.gmp_rep), (mpz_ptr)&(b.gmp_rep) ) ;
  if (g.priv_sign() < 0) { Integer::negin(u); Integer::negin(v); return Integer::negin(g);}
  return g;
}


Integer pp( const Integer& P, const Integer& Q ) 
{
  Integer U = P ; 
  Integer V = gcd(P,Q) ; 
  // -- computes the prime part U of g relatively to U
  while ( V != Integer::one ) 
  {
    U = U / V ;
    V = gcd( U,V) ;
  } 
  return U ;
}
