#ifndef _GIVARO_GFQ1_H_
#define _GIVARO_GFQ1_H_
// ==========================================================================
// file: givgfq.h 
// Time-stamp: <04 Oct 01 18:30:45 Jean-Guillaume.Dumas@imag.fr>
// (c) Givaro Team
// date: 1999
// version: 
// author: Jean-Guillaume.Dumas
// Description:
//   Arithmetic on GF(p^k), with p a prime number less than 2^15
// ==========================================================================

#include "givconfig.h"
#include "givinteger.h"
#include <iostream.h>
#include <vector.h>

  // ------------------------------------------------- class GFqDom

template<class TT> class GFqDom {
protected:
    typedef typename Signed_Trait<TT>::unsigned_type UTT;
    typedef TT Rep;
public:
    typedef GFqDom<TT> Self_t;
    
    
    typedef Rep element;
//     class element {
//     public:
//         mutable Rep _value;

//         element() {}
//     };    

    typedef UTT Residu_t;

        // ----- Representation of vector of the element
    typedef Rep* Array;
    typedef const Rep* constArray;

        // ----- Constantes 
//    const Rep zero;
//    const Rep one;
    Rep zero;
    Rep one;

    GFqDom(): zero(0), _log2pol(0), _pol2log(0),_plus1(0) {}

    GFqDom( const UTT P, const UTT e = 1);

        // Allows to choose the randomization 
        // and therefore the field generator
//     template<class RandIter >
//     GFqDom(RandIter& g, const UTT P, const UTT e = 1);
    ~GFqDom() {};

        // Access to the modulus, characteristic, size, exponent

    UTT residu() const;
    UTT characteristic() const;
    UTT cardinality() const;
    UTT exponent() const;
    UTT size() const;

        // Initialization of elements
    Rep& init( Rep& ) const;
    Rep& init( Rep&, const TT ) const;
    Rep& init( Rep&, const UTT ) const;
    Rep& init( Rep&, const Integer ) const;
    Rep& init( Rep&, const double ) const ;
//    Rep& init( Rep&, const Rep ) const;

        // -- Misc: r <- a mod p
//     Rep& assign(Rep& r, const TT a) const;
//     Rep& assign(Rep& r, const UTT a) const;
    Rep& assign(Rep& r, const Integer a) const;
    Rep& assign(Rep& r, const Rep a) const;
    void assign ( const size_t sz, Array r, constArray a ) const;

        // --- IO methods for the Domain
    istream& read ( istream& s );
    ostream& write( ostream& s ) const;
        // --- IO methods for the elements
    istream& read ( istream& s, Rep& a ) const;
    Rep& read (Rep&, const TT ) const;
    Rep& read (Rep&, const UTT ) const;
    Rep& read (Rep&, const int ) const;
    Rep& read (Rep&, const unsigned int ) const;
    Rep& read (Rep&, const Integer ) const;
    
    ostream& write( ostream& s, const Rep a ) const;
    TT write(const Rep a ) const;
    TT& write(TT&, const Rep a ) const;
    UTT& write(UTT&, const Rep a ) const;
    int& write(int&, const Rep a ) const;
    unsigned int& write(unsigned int&, const Rep a ) const;
    Integer& write(Integer&, const Rep a ) const;

    ostream& convert( ostream& s, const Rep a ) const { return write(s,a); }
    TT convert(const Rep a ) const {return write(a);};
    TT& convert(TT& i, const Rep a ) const {return write(i,a);};
    UTT& convert(UTT& i, const Rep a ) const {return write(i,a);};
    int& convert(int& i, const Rep a ) const {return write(i,a);};
    double& convert(double& i, const Rep a ) const ;
    unsigned int& convert(unsigned int& i, const Rep a ) const {return write(i,a);};
    Integer& convert(Integer& i, const Rep a ) const {return write(i,a);};


        // Test operators
    inline int operator== (const GFqDom<TT>& a) const;
    inline int operator!= (const GFqDom<TT>& a) const;

        // Miscellaneous functions
    bool isequal( const Rep, const Rep  ) const;
    bool isnequal ( const Rep , const Rep ) const;
    bool iszero( const Rep ) const;
    bool isnzero( const Rep ) const;
    bool isone ( const Rep ) const;
    size_t length ( const Rep ) const;

    bool areEqual( const Rep a, const Rep b ) const { return isequal(a,b);  } ;
    bool isZero( const Rep a ) const { return iszero(a); }
    bool isOne ( const Rep a ) const { return isone(a); }


        // ----- Operations with reduction: r <- a op b mod p, r <- op a mod p
    Rep& mul (Rep& r, const Rep a, const Rep b) const;
    Rep& div (Rep& r, const Rep a, const Rep b) const;
    Rep& add (Rep& r, const Rep a, const Rep b) const;
    Rep& sub (Rep& r, const Rep a, const Rep b) const;
    Rep& neg (Rep& r, const Rep a) const;
    Rep& inv (Rep& r, const Rep a) const;

    Rep& mulin (Rep& r, const Rep a) const;
    Rep& divin (Rep& r, const Rep a) const;
    Rep& addin (Rep& r, const Rep a) const;
    Rep& subin (Rep& r, const Rep a) const;
    Rep& negin (Rep& r) const;
    Rep& invin (Rep& r) const;

        // ----- Operations with reduction: r <- a op b mod p, r <- op a mod p
    void mul (const size_t sz, Array r, constArray a, constArray b) const;
    void mul (const size_t sz, Array r, constArray a, Rep b) const;

    void div (const size_t sz, Array r, constArray a, constArray b) const;
    void div (const size_t sz, Array r, constArray a, Rep b) const;

    void add (const size_t sz, Array r, constArray a, constArray b) const;
    void add (const size_t sz, Array r, constArray a, Rep b) const;

    void sub (const size_t sz, Array r, constArray a, constArray b) const;
    void sub (const size_t sz, Array r, constArray a, Rep b) const;
    void neg (const size_t sz, Array r, constArray a) const;
    void inv (const size_t sz, Array r, constArray a) const;

    Rep& axpy (Rep& r, const Rep a, const Rep b, const Rep c) const;
    void axpy (const size_t sz, Array r, Rep a, constArray x, constArray y) const;
    void axpy (const size_t sz, Array r, Rep a, constArray x, Rep c) const;

        // -- axpyin: r <- r + a * x mod p
    Rep& axpyin (Rep& r, const Rep a, const Rep b) const;
    void axpyin (const size_t sz, Array r, Rep a, constArray x) const;

        // -- axmy: r <- a * x - y mod p
    Rep& axmy (Rep& r, const Rep a, const Rep b, const Rep c) const;
    void axmy (const size_t sz, Array r, Rep a, constArray x, constArray y) const;
    void axmy (const size_t sz, Array r, Rep a, constArray x, Rep c) const;

        // -- amxy: r <- c - a * b mod p
    Rep& amxy (Rep& r, const Rep a, const Rep b, const Rep c) const;

        // -- axmyin: r <- r - a * b mod p
    Rep& axmyin (Rep& r, const Rep a, const Rep b) const;
    void axmyin (const size_t sz, Array r, Rep a, constArray x) const;

//   // -- sqpyin: r <- r + a * a mod p
//     Rep& sqpyin (Rep& r, const Rep a) const;
    

        // -- axpyin: r <- r - a * b mod p
    Rep& amxyin (Rep& r, const Rep a, const Rep b) const;

        // <- \sum_i a[i], return 1 if a.size() ==0,
    void reduceadd ( Rep& r, const size_t sz, constArray a ) const; 

        // <- \prod_i a[i], return 1 if a.size() ==0,
    void reducemul ( Rep& r, const size_t sz, constArray a ) const; 

        // <- \sum_i a[i] * b[i] 
    Rep& dotprod ( Rep& r, const size_t sz, constArray a, constArray b ) const; 

        // ----- random generators
   // ----- random generators

    template<class RandIter> Rep& random(RandIter& g, Rep& r) const ;
    template<class RandIter> Rep& random(RandIter& g, Rep& r, long s) const ;
    template<class RandIter> Rep& random(RandIter& g, Rep& r, const Rep& b) const ;
    template<class RandIter> Rep& nonzerorandom(RandIter& g, Rep& r) const ;
    template<class RandIter> Rep& nonzerorandom(RandIter& g, Rep& r, long s) const ;
    template<class RandIter> Rep& nonzerorandom(RandIter& g, Rep& r, const Rep& b) const ;
    

//         // - Set to a non zero random value
//     void set_nrand(Rep&) const;
//         // - Set to a random value
//     void set_rand(Rep&) const;
    
#ifdef __GIVARO_COUNT__
    void clear() {
        _add_count = 0;
        _mul_count = 0;
        _neg_count = 0;
        _div_count = 0;
        _sub_count = 0;
        _inv_count = 0;    
        _add_call = 0;
        _mul_call = 0;
        _neg_call = 0;
        _div_call = 0;
        _sub_call = 0;
        _inv_call = 0;    
    }
    
    void info() const {
        cerr << "Mul Call: " << _mul_call << ", real: " << _mul_count << endl;
        cerr << "Add Call: " << _add_call << ", real: " << _add_count << endl;
        cerr << "Div Call: " << _div_call << ", real: " << _div_count << endl;
        cerr << "Sub Call: " << _sub_call << ", real: " << _sub_count << endl;
        cerr << "Neg Call: " << _neg_call << ", real: " << _neg_count << endl;
        cerr << "Inv Call: " << _inv_call << ", real: " << _inv_count << endl;
    }
#endif
    

protected:
//    const UTT _characteristic;
//    const UTT _exponent;
//    const UTT _q;
//    const UTT _qm1;
//    const UTT _qm1o2;
    UTT _characteristic;
    UTT _exponent;
    UTT _q;
    UTT _qm1;
    UTT _qm1o2;

        // G is a generator of GF(q)
        // p is GF(q)'s characteristic
        // log2pol[ i ] = G^i(p)
        // pol2log[ j ] = i such that log2pol[i] = j
        // plus1[i] = k such that G^i + 1 = G^k
    vector<UTT> _log2pol;
    vector<UTT> _pol2log;
    vector<TT> _plus1;


#ifdef __GIVARO_COUNT__
static    long long _add_count;
static    long long _mul_count;
static    long long _neg_count;
static    long long _div_count;
static    long long _sub_count;
static    long long _inv_count;    
static    long long _add_call;
static    long long _mul_call;
static    long long _neg_call;
static    long long _div_call;
static    long long _sub_call;
static    long long _inv_call;    
#endif

    static void Init();
    static void End();
};


#include "givgfq.inl"

#endif 
