/****************************************************************************
**
*W  matrix.h          Homology source                  Dumas & Saunders
**
*H  @(#)$Id: matrix.h,v 0.0 1999/08
**
*Y  Copyright (C)  1999,  Us guys
**
*/

#ifndef _CPP_C_MATRIX_INTERFACE_
#define _CPP_C_MATRIX_INTERFACE_
typedef unsigned long ulong;

#include "commentator.h"

extern "C" {
#include "gmp.h"

typedef int TCount;
typedef int TVectorValue;
typedef struct
{
  TCount Index;
  TVectorValue Val;
} TComponent;

typedef struct
{
  TComponent * Components;
  TVectorValue MaxEl; /* unused */
  TCount Count;
  TCount Size; /* unused */
} TSparseVector;

typedef struct
{
  TSparseVector * Rows;
  TCount NumRows, NumColumns;
} TSparseMatrixArray;

typedef struct
{
  mpz_t Coefficient;
  TCount Count;
} TInvariantFactor;

typedef struct
{
  TInvariantFactor *InvariantFactors;
  TCount Rows, Columns, NumberOfInvariantFactors;
} TSmithForm;


#define hmValenceElim   3
#define hmValenceBB     4
#define SMITHFORM       0
#define INTEGRANK       1
#define TORSIONRANK     2
#define PRIMERANK       4
#define PRIMEPOWERRANK  5
#define CARACTERISTICROWS 6

/**************************************************************************************************
 ** Main handler for C++ code                                                                    **
 **                                                                                              **
 ** Commentator : Message (+ timing ?)  handler                                                  **
 ** Matrix      : A sparse matrix or the null pointer if we use a file.                          **
 ** File_Name   : Garbage or full location of the sparse matrix                                  **
 ** MaxRank     : An upper bound for the rank of the matrix                                      **
 ** Dim         : An indication (if simplicial complex, it is the dimension of the map)          **
 ** Algorithm   : Method of computation                                                          **
 **               - 3  : Valence with elimination                                                **  
 **               - 4  : Valence with black-box                                                  **
 ** Mode        : Type of computation                                                            **  
 **               - 0  : Smith Normal Form                                                       **
 **               - 1  : Integer Rank                                                            **
 **               - 2  : rank N-Torsion                                                          **
 **               - 4  : rank modulo coeff (coeff is prime)                                      **
 **               - 5  : rank modulo coeff (coeff is a prime power)                              **
 ** coeff       : If 'Mode' is 2 then it is the modulus                                          **
 ** SmithForm   : Result depending on 'Mode'                                                     **
 **               - 0  : rows, columns, number of invariants, list of value/exponent             **  
 **               - 1  : rows, columns, 1, 1, rank                                               **
 **               - 2  : rows, columns, 1, list of coeff/exponent                                **
 **               - 4  : rows, columns, 1, coeff, rank                                           **
 **               - 5  : rows, columns, 1, coeff, rank                                           **
 ** error	: Error tolerance (must be strictly between 0 and 1 to be taken into account)         **
 **		  - 0  : default value --> let the algorithm decide				             **
 **		  - e  : 0 < e < 1     --> Results will be correct with probability at least e        **
 **		  - >=1: impossible    --> let the algorithm decide				             **
 **************************************************************************************************/

void MatrixSmithFormValence (Commentator *Comm,
                             const TSparseMatrixArray * Matrix,
                             char * File_Name,
                             TCount MaxRank, TCount Dim,
                             int Algorithm, int Mode, const mpz_t * coeff,
                             TSmithForm * SmithForm,
			     char * numerator, char * denominator );
 
}


// -------------------------------------------------
// C++ call
// -------------------------------------------------
#include "givinteger.h" 
#include "LinBox/lin_invariants.h" 

void MatrixSmithFormValenceCPP (Commentator & Comm,
                                long ni, long nj,
                                char * File_Name,
                                int Method, int Mode, Integer& R,
                                Invariants<ulong>& Inv,
				char * numerator, char * denominator ) ;

#endif 
